import os
import pickle
import random
import string

import click
import numpy as np
from joblib import Parallel, delayed

import scphylo as scp
from scphylo.tl.partition_function._pf import draw_sample_clt


@click.command(short_help="Get samples or calculate for PartF.")
# @click.argument(
#     "kind",
#     required=True,
#     type=click.Choice(["samples", "calculate"]),
# )
@click.argument(
    "genotype_file",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
@click.argument(
    "alpha",
    required=True,
    type=float,
)
@click.argument(
    "beta",
    required=True,
    type=float,
)
@click.option(
    "--n_samples",
    default=1000,
    type=int,
    show_default=True,
    help="Number of samples.",
)
@click.option(
    "--n_threads",
    default=8,
    type=int,
    show_default=True,
    help="Number of threads.",
)
def partf(genotype_file, alpha, beta, n_samples, n_threads):
    """Get samples for calculating the partition function.

    scphylo partf input.SC 0.0001 0.1
    """
    scp.settings.verbosity = "info"

    df_input = scp.io.read(genotype_file)
    outfile = os.path.splitext(genotype_file)[0]
    scp.ul.mkdir(f"{outfile}.partf.samples")

    I_mtr = df_input.values
    t1 = I_mtr * (1 - beta) / (alpha + 1 - beta)
    t2 = (1 - I_mtr) * beta / (beta + 1 - alpha)
    P = t1 + t2
    P[I_mtr == 3] = 0.5
    P = P.astype(np.float64)

    edges_list = []
    subtrees_list = []
    tree_our_prob_list = []

    def run():
        return draw_sample_clt(P, False, c=1, coef=10)

    output = Parallel(n_jobs=n_threads)(delayed(run)() for i in range(0, n_samples))

    for edges, subtrees, prior_prob in output:
        edges_list.append(edges)
        subtrees_list.append(subtrees)
        tree_our_prob_list.append(prior_prob)
    samples_object = (edges_list, subtrees_list, tree_our_prob_list)
    filename = "".join(random.choice(string.ascii_letters) for _ in range(16))
    filename = f"{outfile}.partf.samples/{filename}.pkl"
    with open(filename, "wb") as f:
        pickle.dump(samples_object, f)

    return None



"""Tools Module."""

from scphylo.tl.cna import infercna
from scphylo.tl.consensus import consensus, consensus_day
from scphylo.tl.fitch import fitch
from scphylo.tl.partition_function import partition_function
from scphylo.tl.score import ad, caset, cc, disc, dl, gs, mltd, mp3, rf, tpted
from scphylo.tl.solver import (
    bnb,
    booster,
    cardelino,
    cellphy,
    dendro,
    gpps,
    grmt,
    huntress,
    infscite,
    iscistree,
    onconem,
    phiscs_readcount,
    phiscsb,
    phiscsb_bulk,
    phiscsi,
    phiscsi_bulk,
    rscistree,
    sbm,
    scelestial,
    sciphi,
    scistree,
    scite,
    siclonefit,
    sphyr,
)

__all__ = (
    infercna,
    consensus,
    consensus_day,
    partition_function,
    sbm,
    ad,
    cc,
    dl,
    mltd,
    tpted,
    bnb,
    booster,
    cardelino,
    dendro,
    huntress,
    infscite,
    iscistree,
    onconem,
    phiscsi_bulk,
    phiscs_readcount,
    phiscsb,
    phiscsb_bulk,
    phiscsi,
    rscistree,
    scistree,
    scite,
    siclonefit,
    fitch,
    caset,
    disc,
    mp3,
    rf,
    gs,
    sphyr,
    grmt,
    sciphi,
    gpps,
    scelestial,
    cellphy,
)



"""Methyl Module."""

from collections import OrderedDict

import click

from scphylo.commands.methyl._builds import build_crcs


class NaturalOrderGroup(click.Group):
    """Command group trying to list subcommands in the order they were added.

    Make sure you initialize the `self.commands` with OrderedDict instance.
    With decorator, use::
        @click.group(cls=NaturalOrderGroup, commands=OrderedDict())
    """

    def list_commands(self, ctx):
        """List command names as they are in commands dict.

        If the dict is OrderedDict, it will preserve the order commands
        were added.
        """
        return self.commands.keys()


@click.group(
    name="methyl",
    cls=NaturalOrderGroup,
    commands=OrderedDict(),
    context_settings={"max_content_width": 300, "terminal_width": 300},
)
def cli_methyl():
    """Methyl module."""
    return None


cli_methyl.add_command(build_crcs)



import math
import random

import anndata as ad
import numpy as np
import pandas as pd

import scphylo as scp


def simulate(n_cells=10, n_muts=10, n_clones=3, alpha=0.00001, beta=0.1, missing=0):
    """Simulate single-cell noisy genotype matrix.

    This function is using :cite:`OncoNEM`.

    Parameters
    ----------
    n_cells : :obj:`int`, optional
        Number of cells, by default 10
    n_muts : :obj:`int`, optional
        Number of mutations, by default 10
    n_clones : :obj:`int`, optional
        Number of clones, by default 3
    alpha : :obj:`float`, optional
        False positive rate, by default 0.00001
    beta : :obj:`float`, optional
        False negative rate, by default 0.1
    missing : :obj:`int`, optional
        Missing entry rate, by default 0

    Returns
    -------
    :class:`pandas.DataFrame`
        A genotype matrix where 0 is absent, 1 is present and 3 is missing.
    """
    # TODO: replace

    onconem, onconem_is_not_imported = scp.ul.import_rpy2(
        "oncoNEM",
        "BiocManager::install('graph')\n"
        "devtools::install_bitbucket('edith_ross/oncoNEM')\n",
    )
    if onconem_is_not_imported:
        scp.logg.error("Unable to import a package!")

    import rpy2.robjects as ro
    from rpy2.robjects import pandas2ri

    dat = onconem.simulateData(
        N_cells=n_cells,
        N_normalContam=0,
        N_clones=n_clones,
        N_unobs=0,
        N_sites=n_muts,
        FPR=alpha,
        FNR=beta,
        p_missing=missing,
        randomizeOrder=False,
    )

    with ro.conversion.localconverter(ro.default_converter + pandas2ri.converter):
        dat = ro.conversion.rpy2py(dat.rx2("D"))
    dat[dat == 2] = 3
    df = pd.DataFrame(dat.T, dtype=int)
    df.columns = [f"mut{x}" for x in df.columns]
    df.index = [f"cell{x}" for x in df.index]

    return df


def add_noise(df_in, alpha, beta, missing):
    """Add noise to the input genotype matrix.

    These noise includes:
    1) False positive errors (alpha)
    2) False negative errors (beta)
    3) Missing entry errors (missing)

    Parameters
    ----------
    df_in : :class:`pandas.DataFrame`
        Input genotype matrix.
    alpha : :obj:`float`
        False positive error rate.
    beta : :obj:`float`
        False negative error rate.
    missing : :obj:`float`
        Missing entry error rate.

    Returns
    -------
    :class:`pandas.DataFrame`
        A noisy genotype matrix where 0 is absent, 1 is present and 3 is missing.
    """

    def toss(p):
        return True if random.random() < p else False

    data = df_in.values
    n, m = df_in.shape
    data2 = -1 * np.ones(shape=(n, m)).astype(int)
    countFP = 0
    countFN = 0
    countNA = 0
    countOneZero = 0
    indexNA = []
    changedBefore = []
    for i in range(n):
        for j in range(m):
            indexNA.append([i, j])
            countOneZero = countOneZero + 1
    random.shuffle(indexNA)
    nas = math.ceil(countOneZero * missing)
    for i in range(int(nas)):
        [a, b] = indexNA[i]
        changedBefore.append([a, b])
        data2[a][b] = 3
        countNA = countNA + 1
    for i in range(n):
        for j in range(m):
            if data2[i][j] != 3:
                if data[i][j] == 1:
                    if toss(beta):
                        data2[i][j] = 0
                        countFN = countFN + 1
                    else:
                        data2[i][j] = data[i][j]
                elif data[i][j] == 0:
                    if toss(alpha):
                        data2[i][j] = 1
                        countFP = countFP + 1
                    else:
                        data2[i][j] = data[i][j]
                else:
                    scp.logg.error("Wrong Input")

    # scp.logg.info(f"FNs={countFN}, FPs={countFP}, NAs={countNA}")

    df_out = pd.DataFrame(data2)
    df_out.columns = df_in.columns
    df_out.index = df_in.index
    df_out.index.name = "cellIDxmutID"

    return df_out


def add_doublets(df_ground, df_noisy, alpha, beta, missing, doublet):
    df_doublet = df_noisy.copy()
    doublet_cells = []
    for _ in range(int(doublet * df_ground.shape[0])):
        r1 = np.random.choice(df_ground.index, replace=False, size=1)
        while r1 in doublet_cells:
            r1 = np.random.choice(df_ground.index, replace=False, size=1)
        doublet_cells.append(r1)
        r2 = np.random.choice(df_ground.index, replace=False, size=1)
        df_doublet.loc[r1] = 1 * np.logical_or(df_ground.loc[r1], df_ground.loc[r2])
        df_doublet.loc[r1] = scp.datasets.add_noise(
            df_doublet.loc[r1], alpha=alpha, beta=beta, missing=missing
        )
    return df_doublet


def add_readcount(df_in, mean_coverage=50, seed=None):
    """Add readcount to the input genotype matrix.

    Parameters
    ----------
    df_in : :class:`pandas.DataFrame`
        Input genotype matrix.
    mean_coverage : :obj:`int`, optional
        The average coverage, by default 50
    seed : :obj:`int`, optional
        The seed for numpy random function, by default None

    Returns
    -------
    adata : :class:`anndata.AnnData`
        The genotype data with readcount info contains layers of mutant and total.
    """
    if seed is not None:
        np.random.seed(seed)
    n, m = df_in.shape
    I_mtx = df_in.values
    T_mtx = np.random.binomial(n=mean_coverage, p=0.5, size=(n, m))

    R_mtx = np.zeros(I_mtx.shape, dtype=int)
    for i in range(I_mtx.shape[0]):
        for j in range(I_mtx.shape[1]):
            if I_mtx[i, j] == 0:
                # rr allele
                R_mtx[i, j] = np.ceil(T_mtx[i, j] * np.random.beta(20, 1))
            elif I_mtx[i, j] == 1:
                if np.random.uniform(0, 1) < 0.5:
                    # aa allele
                    R_mtx[i, j] = np.ceil(T_mtx[i, j] * np.random.beta(1, 20))
                else:
                    # ra allele
                    R_mtx[i, j] = np.ceil(T_mtx[i, j] * np.random.beta(20, 20))
            elif I_mtx[i, j] == 3:
                R_mtx[i, j] = 0
                T_mtx[i, j] = 0
            else:
                scp.logg.error("Wrong Input")
    adata = ad.AnnData(df_in, dtype=int)
    adata.layers["total"] = T_mtx
    adata.layers["mutant"] = T_mtx - R_mtx
    return adata



import contextlib
import datetime
import functools
import multiprocessing
import os
import shutil
import tempfile
import time

import joblib
import numpy as np
import pkg_resources

import scphylo as scp


def log_input(df_in):
    size = df_in.shape[0] * df_in.shape[1]
    scp.logg.info(f"input -- size: {df_in.shape[0]}x{df_in.shape[1]}")
    scp.logg.info(
        f"input -- 0: {np.sum(df_in.values == 0)}#,"
        f" {100*np.sum(df_in.values == 0)/size:.1f}%"
    )
    scp.logg.info(
        f"input -- 1: {np.sum(df_in.values == 1)}#,"
        f" {100*np.sum(df_in.values == 1)/size:.1f}%"
    )
    scp.logg.info(
        f"input -- NA: {np.sum(df_in.values == 3)}#,"
        f" {100*np.sum(df_in.values == 3)/size:.1f}%"
    )
    scp.logg.info(f"input -- CF: {is_conflict_free_gusfield(df_in)}")


def log_output(df_out, running_time):
    size = df_out.shape[0] * df_out.shape[1]
    scp.logg.info(f"output -- size: {df_out.shape[0]}x{df_out.shape[1]}")
    scp.logg.info(
        f"output -- 0: {np.sum(df_out.values == 0)}#,"
        f" {100*np.sum(df_out.values == 0)/size:.1f}%"
    )
    scp.logg.info(
        f"output -- 1: {np.sum(df_out.values == 1)}#,"
        f" {100*np.sum(df_out.values == 1)/size:.1f}%"
    )
    scp.logg.info(
        f"output -- NA: {np.sum(df_out.values == 3)}#,"
        f" {100*np.sum(df_out.values == 3)/size:.1f}%"
    )
    icf = is_conflict_free_gusfield(df_out)
    scp.logg.info("output -- CF: ", end="")
    if icf:
        scp.logg.info(icf, color="green")
    else:
        scp.logg.info(icf, color="red")
    scp.logg.info(
        f"output -- time: {running_time:.1f}s"
        f" ({datetime.timedelta(seconds=running_time)})"
    )


def log_flip(df_in, df_out):
    flips_0_1, flips_1_0, flips_na_0, flips_na_1 = count_flips(
        df_in.values, df_out.values, 3
    )
    fn_rate, fp_rate, na_rate = infer_rates(df_in.values, df_out.values, 3)
    scp.logg.info(f"flips -- #0->1: {flips_0_1}")
    scp.logg.info(f"flips -- #1->0: {flips_1_0}")
    scp.logg.info(f"flips -- #NA->0: {flips_na_0}")
    scp.logg.info(f"flips -- #NA->1: {flips_na_1}")
    scp.logg.info(f"rates -- FN: {fn_rate:.3f}")
    scp.logg.info(f"rates -- FP: {fp_rate:.8f}")
    scp.logg.info(f"rates -- NA: {na_rate:.3f}")


def calc_nll_matrix(df_in, df_out, alpha, beta):
    if alpha == 0 or beta == 0:
        return None
    columns = np.intersect1d(df_in.columns, df_out.columns)
    indices = np.intersect1d(df_in.index, df_out.index)
    D = df_in.loc[indices, columns].values
    E = df_out.loc[indices, columns].values
    removedMutations = []
    objective = 0
    for j in range(len(columns)):
        numZeros = 0
        numOnes = 0
        for i in range(len(indices)):
            if D[i, j] == 0:
                numZeros += 1
                objective += np.log(beta / (1 - alpha)) * E[i, j]
            elif D[i, j] == 1:
                numOnes += 1
                objective += np.log((1 - beta) / alpha) * E[i, j]
        objective += numZeros * np.log(1 - alpha)
        objective += numOnes * np.log(alpha)
        if j in removedMutations:
            objective -= numZeros * np.log(1 - alpha) + numOnes * (
                np.log(alpha) + np.log((1 - beta) / alpha)
            )
    return -objective


def stat(df_in, df_out, alpha, beta, running_time):
    log_input(df_in)
    log_output(df_out, running_time)
    log_flip(df_in, df_out)
    nll = calc_nll_matrix(df_in, df_out, alpha, beta)
    scp.logg.info(f"score -- NLL: {nll}")


def parse_params_file(filename):
    def _parse_params_file_helper(param):
        try:
            value = basename.split(f"{param}_")[1]
            if "-" in value:
                value = value.split("-")[0]
            else:
                value = value.split(".")[0]
            return float(value) if "." in value else int(value)
        except IndexError:
            return None

    data = {}
    _, basename = dir_base(filename)
    for param in [
        "simNo",
        "s",
        "m",
        "h",
        "minVAF",
        "ISAV",
        "n",
        "fp",
        "fn",
        "na",
        "d",
        "l",
    ]:
        value = _parse_params_file_helper(param)
        if value is not None:
            data[param] = value
    return data


def parse_log_file(filename):
    result = {}
    _, basename = dir_base(filename)
    result["tool"] = basename.split(".")[-1]
    with open(filename) as fin:
        for line in fin:
            line = line.strip()
            if "output -- time: " in line:
                result["running_time"] = float(
                    line.replace("output -- time: ", "").split()[0].replace("s", "")
                )
            if "output -- CF: " in line:
                result["is_cf"] = bool(line.replace("output -- CF: ", "").split()[-1])
            if "rates -- FN: " in line:
                result["fn_rate"] = float(line.replace("rates -- FN: ", "").split()[-1])
            if "rates -- FP: " in line:
                result["fp_rate"] = float(line.replace("rates -- FP: ", "").split()[-1])
    return result


def parse_score_file(filename):
    result = {}
    _, basename = dir_base(filename)
    result["tool"] = basename.split(".")[-1]
    with open(filename) as fin:
        for line in fin:
            line = line.strip()
            result[line.split("=")[0]] = float(line.split("=")[1])
    return result


def count_flips(I_mtr, O_mtr, na_value=3):
    flips_0_1 = 0
    flips_1_0 = 0
    flips_na_0 = 0
    flips_na_1 = 0
    n, m = I_mtr.shape
    for i in range(n):
        for j in range(m):
            if I_mtr[i, j] == 0 and O_mtr[i, j] == 1:
                flips_0_1 += 1
            elif I_mtr[i, j] == 1 and O_mtr[i, j] == 0:
                flips_1_0 += 1
            elif I_mtr[i, j] == na_value and O_mtr[i, j] == 0:
                flips_na_0 += 1
            elif I_mtr[i, j] == na_value and O_mtr[i, j] == 1:
                flips_na_1 += 1
    return flips_0_1, flips_1_0, flips_na_0, flips_na_1


def infer_rates(I_mtr, O_mtr, na_value=3):
    flips_0_1, flips_1_0, flips_na_0, flips_na_1 = count_flips(I_mtr, O_mtr, na_value)
    fn_rate = flips_0_1 / ((O_mtr == 1) & (I_mtr != na_value)).sum()
    fp_rate = flips_1_0 / ((O_mtr == 0) & (I_mtr != na_value)).sum()
    na_rate = (flips_na_1 + flips_na_0) / I_mtr.size
    return fn_rate, fp_rate, na_rate


def is_conflict_free(df_in):
    """Check conflict-free criteria via Gusfield algorithm.

    The order of this algorithm is :math:`O(nm^2)`
    where n is the number of cells and m is the number of mutations.

    Parameters
    ----------
    df_in : :class:`pandas.DataFrame`
        Input genotype matrix.

    Returns
    -------
    :obj:`bool`
        A Boolean checking if the input conflict-free or not.

    See Also
    --------
    :func:`scphylo.ul.is_conflict_free_gusfield`.
    """
    D = df_in.astype(int).values
    if not np.array_equal(np.unique(D), [0, 1]):
        return False
    conflict_free = True
    for p in range(D.shape[1]):
        for q in range(p + 1, D.shape[1]):
            oneone = False
            zeroone = False
            onezero = False
            for r in range(D.shape[0]):
                if D[r, p] == 1 and D[r, q] == 1:
                    oneone = True
                if D[r, p] == 0 and D[r, q] == 1:
                    zeroone = True
                if D[r, p] == 1 and D[r, q] == 0:
                    onezero = True
            if oneone and zeroone and onezero:
                conflict_free = False
                return conflict_free
    return conflict_free


def is_conflict_free_gusfield(df_in):
    """Check conflict-free criteria via Gusfield algorithm.

    This is an implementation of algorithm 1.1 in :cite:`Gusfield_1991`.

    The order of this algorithm is :math:`O(nm)`
    where n is the number of cells and m is the number of mutations.

    Parameters
    ----------
    df_in : :class:`pandas.DataFrame`
        Input genotype matrix.

    Returns
    -------
    :obj:`bool`
        A Boolean checking if the input conflict-free or not.

    Examples
    --------
    >>> sc = scp.datasets.test()
    >>> scp.ul.is_conflict_free_gusfield(sc)
    False

    See Also
    --------
    :func:`scphylo.ul.is_conflict_free`.
    """
    I_mtr = df_in.astype(int).values
    if not np.array_equal(np.unique(I_mtr), [0, 1]):
        return False

    def _sort_bin(a):
        b = np.transpose(a)
        b_view = np.ascontiguousarray(b).view(
            np.dtype((np.void, b.dtype.itemsize * b.shape[1]))
        )
        idx = np.argsort(b_view.ravel())[::-1]
        c = b[idx]
        return np.transpose(c), idx

    Ip = I_mtr.copy()
    O_mtr, _ = _sort_bin(Ip)
    Lij = np.zeros(O_mtr.shape, dtype=int)
    for i in range(O_mtr.shape[0]):
        maxK = 0
        for j in range(O_mtr.shape[1]):
            if O_mtr[i, j] == 1:
                Lij[i, j] = maxK
                maxK = j + 1
    Lj = np.amax(Lij, axis=0)
    for i in range(O_mtr.shape[0]):
        for j in range(O_mtr.shape[1]):
            if O_mtr[i, j] == 1:
                if Lij[i, j] != Lj[j]:
                    return False
    return True


def tmpdir(prefix="scphylo.", suffix=".scphylo", dirname="."):
    return tempfile.mkdtemp(suffix=suffix, prefix=prefix, dir=dirname)


def tmpfile(prefix="scphylo.", suffix=".scphylo", dirname="."):
    return tempfile.mkstemp(suffix=suffix, prefix=prefix, dir=dirname)[1]


def tmpdirsys(prefix="scphylo.", suffix=".scphylo", dirname="."):
    return tempfile.TemporaryDirectory(suffix=suffix, prefix=prefix)


def cleanup(dirname):
    shutil.rmtree(dirname)


def remove(filename):
    if os.path.exists(filename):
        os.remove(filename)


def dir_base(infile):
    basename = os.path.splitext(os.path.basename(infile))[0]
    dirname = os.path.dirname(infile)
    return dirname, basename


def dirbase(infile):
    return os.path.splitext(infile)[0]


def mkdir(indir):
    if not os.path.exists(indir):
        os.makedirs(indir)
    return indir


def executable(binary, appname):
    executable = shutil.which(binary)
    if executable is None:
        if not os.path.exists(f"{scp.settings.tools}/{binary}"):
            scp.logg.error(
                f"Cannot find the binary file of {appname} with `{binary}` name!"
            )
        else:
            executable = f"{scp.settings.tools}/{binary}"
    return executable


def timeit(f):
    def wrap(*args, **kwargs):
        start_time = time.time()
        ret = f(*args, **kwargs)
        end_time = time.time()
        if end_time - start_time < 60:
            scp.logg.info(f"Time needed for {f.__name__}: {end_time - start_time:.3f}")
        else:
            scp.logg.info(
                f"Time needed for {f.__name__}:"
                f" {time.strftime('%Hh:%Mm:%Ss', time.gmtime(end_time - start_time))}"
            )
        return ret

    return wrap


def get_file(key):
    components = key.split("/")
    return pkg_resources.resource_filename(components[0], "/".join(components[1:]))


def with_timeout(timeout):
    def decorator(decorated):
        @functools.wraps(decorated)
        def inner(*args, **kwargs):
            pool = multiprocessing.pool.ThreadPool(1)
            async_result = pool.apply_async(decorated, args, kwargs)
            try:
                return async_result.get(timeout)
            except multiprocessing.TimeoutError:
                return None

        return inner

    return decorator


@contextlib.contextmanager
def tqdm_joblib(tqdm_object):
    class TqdmBatchCompletionCallback(joblib.parallel.BatchCompletionCallBack):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def __call__(self, *args, **kwargs):
            tqdm_object.update(n=self.batch_size)
            return super().__call__(*args, **kwargs)

    old_batch_callback = joblib.parallel.BatchCompletionCallBack
    joblib.parallel.BatchCompletionCallBack = TqdmBatchCompletionCallback
    try:
        yield tqdm_object
    finally:
        joblib.parallel.BatchCompletionCallBack = old_batch_callback
        tqdm_object.close()


def split_mut(mut):
    try:
        ref = mut.split(".")[-2]
        pos = mut.split(".")[-3]
        chrom = mut.split(".")[-4]
        gene = mut.split(".chr")[0].split("_")[1]
        ens = mut.split(".chr")[0].split("_")[0]
        alt = mut.split(".")[-1]
        return ens, gene, chrom, pos, ref, alt
    except IndexError:
        return None, None, None, None, None, None



import glob
import subprocess

import click
import pandas as pd

import scphylo as scp
from scphylo.ul._servers import cmd, write_cmds_get_main


@click.command(short_help="Run RSEM.")
@click.argument(
    "outdir",
    required=True,
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, readable=True, resolve_path=True
    ),
)
@click.argument(
    "ref",
    required=True,
    type=click.Choice(scp.settings.refs),
)
@click.argument(
    "single_or_paired",
    required=True,
    type=click.Choice(["single", "paired"]),
)
@click.option(
    "--time",
    default=["1:00:00", "10:00:00"],
    multiple=True,
    show_default=True,
    help="Time List [indexing, calling].",
)
@click.option(
    "--mem",
    default=["20", "20"],
    multiple=True,
    show_default=True,
    help="Memory List [indexing, calling].",
)
@click.option(
    "--afterok",
    default=None,
    type=str,
    show_default=True,
    help="Afterok.",
)
def rsem(outdir, ref, single_or_paired, time, mem, afterok):
    """Run RSEM.

    scphylo caller rsem /path/to/in/dir hg19|hg38|mm10 paired

    BAM files (*.transcript.bam) --> AnnData file (_expression.h5ad.gz)
    """
    if ref == "hg19":
        config = scp.settings.hg19
    elif ref == "mm10":
        config = scp.settings.mm10
    elif ref == "hg38":
        config = scp.settings.hg38
    else:
        config = None

    def step1(afterok):
        def get_command():
            cmds = ""
            cmds += cmd([f"mkdir -p {outdir}/_indexing/rsem"])
            cmds += cmd([f"module load {scp.settings.tools['rsem']}"])
            cmds += cmd(
                [
                    "rsem-prepare-reference",
                    "--num-threads 32",
                    "--gtf",
                    f"{config['annot']}",
                    f"{config['ref']}",
                    f"{outdir}/_indexing/rsem/rsem",
                ]
            )
            cmds += cmd(["echo Done!"], islast=True)
            return cmds

        cmds_tmp = pd.DataFrame()
        cmds_tmp["cmd"] = [get_command()]
        cmdmain = write_cmds_get_main(
            cmds_tmp,
            "rsem-1of3",
            time[0],
            mem[0],
            None,
            32,
            scp.settings.tools["email"],
            f"{outdir}/_tmp",
            afterok,
        )
        return cmdmain

    def step2(afterok):
        def get_command(sample):
            cmds = ""
            cmds += cmd([f"module load {scp.settings.tools['rsem']}"])
            if single_or_paired == "paired":
                cmds += cmd(
                    [
                        "rsem-calculate-expression",
                        "--bam",
                        "--no-bam-output",
                        "--paired-end",
                        "--estimate-rspd",
                        f"{outdir}/{sample}.transcript.bam",
                        f"{outdir}/_indexing/rsem/rsem",
                        f"{outdir}/{sample}",
                    ]
                )
            else:
                cmds += cmd(
                    [
                        "rsem-calculate-expression",
                        "--bam",
                        "--no-bam-output",
                        "--estimate-rspd",
                        f"{outdir}/{sample}.transcript.bam",
                        f"{outdir}/_indexing/rsem/rsem",
                        f"{outdir}/{sample}",
                    ]
                )
            cmds += cmd(
                [
                    "rm -rf",
                    f"{outdir}/{sample}.stat",
                ]
            )
            cmds += cmd(["echo Done!"], islast=True)
            return cmds

        df_cmds["cmd"] = df_cmds.apply(lambda x: get_command(x["sample"]), axis=1)
        cmdmain = write_cmds_get_main(
            df_cmds,
            "rsem-2of3",
            time[1],
            mem[1],
            None,
            1,
            scp.settings.tools["email"],
            f"{outdir}/_tmp",
            afterok,
        )
        return cmdmain

    def step3(afterok):
        def get_command():
            cmds = ""
            cmds += cmd([f"python {scp.ul.get_file('scphylo.ul/rsem.py')} {outdir}"])
            cmds += cmd(["echo Done!"], islast=True)
            return cmds

        cmds_tmp = pd.DataFrame()
        cmds_tmp["cmd"] = [get_command()]
        cmdmain = write_cmds_get_main(
            cmds_tmp,
            "rsem-3of3",
            time[0],
            mem[0],
            None,
            2,
            scp.settings.tools["email"],
            f"{outdir}/_tmp",
            afterok,
        )
        return cmdmain

    files = glob.glob(f"{outdir}/*.transcript.bam")
    temp = []
    for file in files:
        file = file.split("/")[-1].replace(".transcript.bam", "")
        temp.append({"sample": file})
    df_cmds = pd.DataFrame(temp)

    cmd1 = step1(afterok)
    code1 = subprocess.getoutput(cmd1)
    cmd2 = step2(code1)
    code2 = subprocess.getoutput(cmd2)
    cmd3 = step3(code2)
    code3 = subprocess.getoutput(cmd3)
    scp.logg.info(code3)

    return None



import pandas as pd

import scphylo as scp
from scphylo.ul._trees import _to_newick, to_mtree

human = [
    ("1", 249698942, 125000000),
    ("2", 242508799, 93300000),
    ("3", 198450956, 91000000),
    ("4", 190424264, 50400000),
    ("5", 181630948, 48400000),
    ("6", 170805979, 61000000),
    ("7", 159345973, 59900000),
    ("8", 145138636, 45600000),
    ("9", 138688728, 49000000),
    ("10", 133797422, 40200000),
    ("11", 135186938, 53700000),
    ("12", 133275309, 35800000),
    ("13", 114364328, 17900000),
    ("14", 108136338, 17600000),
    ("15", 102439437, 19000000),
    ("16", 92211104, 36600000),
    ("17", 83836422, 24000000),
    ("18", 80373285, 17200000),
    ("19", 58617616, 26500000),
    ("20", 64444167, 27500000),
    ("21", 46709983, 13200000),
    ("22", 51857516, 14700000),
    ("X", 156040895, 60600000),
    ("Y", 57264655, 12500000),
]
mouse = [
    ("1", 196283350, 125000000),
    ("2", 182113224, 93300000),
    ("3", 160039680, 91000000),
    ("4", 157219549, 50400000),
    ("5", 153573022, 48400000),
    ("6", 149736546, 61000000),
    ("7", 145617427, 59900000),
    ("8", 129401213, 45600000),
    ("9", 124595110, 49000000),
    ("10", 130694993, 40200000),
    ("11", 122082543, 53700000),
    ("12", 120129022, 35800000),
    ("13", 120421639, 17900000),
    ("14", 124902244, 17600000),
    ("15", 104043685, 19000000),
    ("16", 98207768, 36600000),
    ("17", 94987271, 24000000),
    ("18", 90702639, 17200000),
    ("19", 61431566, 26500000),
    ("X", 171368232, 60600000),
    ("Y", 92500857, 12500000),
]


def _get_tree(
    newick,
    line_size,
    label_color,
    tiplab_size,
    inner_node_type,
    inner_node_size,
    distance_labels_to_bottom,
):
    cmd = f"""
    p <- ggtree(read.tree(text='{newick}'),
                layout='dendrogram',
                size={line_size}) %<+% info1 %<+% info2 +
    # geom_tippoint(aes(color={label_color}), size={tiplab_size}) +
    geom_tiplab(aes(color={label_color}),
                align=TRUE,
                linesize={line_size/4},
                angle=-90,
                hjust=-0.1,
                size={tiplab_size},
                fontface='bold') +
    scale_color_identity(guide='none') +
    geom_label2(aes(x=branch, subset=!isTip,
                    label={inner_node_type.lower()}_label),
                fill='white',
                color='black',
                # parse=TRUE,
                size={inner_node_size},
                label.padding=unit(0.1,'lines')) +
    xlim_tree({distance_labels_to_bottom})
    """
    return cmd


def _add_barplot(column_name, column_color, height=0.4):
    cmd = f"""
    p2 <- ggplot(data=info1) +
    geom_bar(aes(x=id_index, y={column_name}, fill={column_color}),
                 width=0.85, stat='identity', show.legend=FALSE) +
    scale_fill_gradient2(low='#E41A1C', midpoint=0, mid='white', high='#377EB8') +
    # scale_fill_identity(guide='none') +
    # scale_color_identity(guide='none') +
    theme_cowplot() +
    theme(axis.title.x=element_blank(),
          axis.text.x=element_blank(),
          axis.ticks.x=element_blank(),
          legend.title=element_text(size=6),
          legend.text=element_text(size=5),
          axis.title.y=element_text(size=6, face='bold'),
          axis.text.y=element_text(size=5),
          axis.line=element_line(size=0.2),
          axis.ticks=element_line(size=0.1),
          axis.ticks.length=unit(0.02, "in"),
    ) +
    labs(y='{column_name}', x='')
    p <- p %>% insert_bottom(p2, height={height})
    """
    return cmd


def _add_chromplot_helper(name, length):
    cmd = f"""
    p2 <- ggplot(data=info1) +
    geom_rect(aes(xmin=id_index-0.3, xmax=id_index+0.3,
                  ymax={length/1000000}, ymin=0, fill='white', color='black'),
              show.legend=FALSE) +
    scale_fill_identity(guide='none') +
    scale_color_identity(guide='none') +
    theme_cowplot() +
    theme(axis.title.x=element_blank(),
          axis.line.x=element_blank(),
          axis.text.x=element_blank(),
          axis.ticks.x=element_blank(),
          legend.title=element_text(size=6),
          legend.text=element_text(size=5),
          axis.title.y=element_text(size=6, face='bold', angle=0, vjust=0.6, hjust=2),
          axis.text.y=element_blank(),
          axis.line.y=element_blank(),
          axis.ticks.y=element_blank(),
          axis.ticks.length=unit(0.02, "in"),
    ) +
    labs(y='chr{name}', x='')
    p2 <- p2 + geom_segment(x=4-0.3, y=2, xend=4+0.3, yend=2,
                            size=0.1, color='blue')
    p <- p %>% insert_bottom(p2, height={length/500000000})
    """
    return cmd


def _add_chromplot():
    cmd = ""
    for m in mouse:
        cmd += _add_chromplot_helper(m[0], m[1])
    return cmd


def _clonal_cell_mutation_list(tree):
    muts_list = []
    cells_list = []
    for _, v, l in tree.edges.data("label"):
        muts = l.split(tree.graph["splitter_mut"])
        if "––" not in tree.nodes[v]["label"]:
            cells = tree.nodes[v]["label"].split(tree.graph["splitter_cell"])
            for mut in muts:
                muts_list.append({"mut": mut, "Node": f"[{v}]"})
            for cell in cells:
                cells_list.append({"cell": cell, "Node": f"[{v}]"})
        else:
            for mut in muts:
                muts_list.append({"mut": mut, "Node": f"[{v}]"})
    cells_list = pd.DataFrame(cells_list).set_index("cell")
    muts_list = pd.DataFrame(muts_list).set_index("mut")
    return cells_list, muts_list


def _newick_info2_mutation_list(tree):
    tree2 = to_mtree(tree)

    row = []
    for node in tree2.nodes:
        if tree2.in_degree(node) == 0:
            row.append(
                {
                    "newick_label": f"Node{node+1}",
                    "nmuts_label": "root",
                    "nodeid_label": f"[{node+1}]",
                    "both_label": f"[{node+1}]: root",
                }
            )
        else:
            row.append(
                {
                    "newick_label": f"Node{node+1}",
                    "nmuts_label": f"{len(tree2.nodes[node]['label'])}",
                    "nodeid_label": f"[{node+1}]",
                    "both_label": f"[{node+1}]: {len(tree2.nodes[node]['label'])}",
                }
            )
    info2 = pd.DataFrame(row)

    row = []
    for node in tree2.nodes:
        if tree2.in_degree(node) == 0:
            continue
        for mut in tree2.nodes[node]["label"]:
            ens, gene, chrom, pos, ref, alt = scp.ul.split_mut(mut)
            if ens is None:
                row.append(
                    {
                        "index": mut,
                        "node_id": f"[{node+1}]",
                    }
                )
            else:
                row.append(
                    {
                        "index": mut,
                        "Node": f"[{node+1}]",
                        "Ensemble": ens,
                        "Gene": gene,
                        "Chrom": chrom,
                        "Position": pos,
                        "Reference": ref,
                        "Alteration": alt,
                    }
                )
    mutation_list = pd.DataFrame(row)
    newick = _to_newick(tree)
    return newick, info2, mutation_list



import networkx as nx


def to_png(tree, filepath, dpi=150):
    mygraph = nx.drawing.nx_agraph.to_agraph(tree)
    mygraph.graph_attr["dpi"] = dpi
    mygraph.layout(prog="dot")
    mygraph.draw(filepath)



from decimal import Decimal

import numpy as np
from tqdm import tqdm

from scphylo.tl.partition_function._clt_sampler import draw_sample_clt


def cell_lineage_tree_prob(P, subtrees):
    r"""Calculate Prob_{A\sim P}[A\in T] in O(n m^2).

    :param P:
    :param subtrees: cell lineage tree
    :return: Probability of this tree in the original distribution( based on P)
    """
    return_value = Decimal(1)
    for j in range(P.shape[1]):
        temp = Decimal(0)
        col = P[:, j]
        for v in subtrees:
            temp += Decimal(np.prod(col * v + (1 - col) * (1 - v)))
        return_value *= temp
    return return_value


def pf_cond_on_one_tree(P, subtrees, cond_c, cond_m):
    r"""Prob_{A\sim P}[\subtree(c, R, A)\cap A\in G| A\in T] in O(n^2).

    :param P:
    :param subtrees: cell lineage tree  n x (2n+1)
    :param cond_c: set of cells
    :param cond_m: one mutation
    :return: conditioned on the given tree what is probability of the given partition
        based on P numerator, denominator are returned separately here
    """
    denominator = Decimal(0)
    numerator = Decimal(0)
    col = P[:, cond_m]
    for v in subtrees:
        prob = Decimal(np.prod(col * v + (1 - col) * (1 - v)))
        denominator += prob
        if np.array_equal(v, cond_c):
            numerator = prob
    return numerator, denominator


def get_samples(P, n_samples, disable_tqdm=True):
    r"""
    N_s *.

    :param P:
    :param n_samples:
    :return: `n_samples` of cell lineage trees.
        The function returns three lists with this length:
        edges_list: for each sample: (n-1) wedges. Each wedge
            [parent, left_child, right_child]
        subtrees_list: for each sample: 2n+1 number of cell subset
            masks, i.e., potential column.
        tree_our_prob_list: for each sample: the probability of us sampling
            the tree. Prob_{T\sim E}[T]
    """
    P = P.astype(np.float64)
    edges_list = []
    subtrees_list = []
    tree_our_prob_list = []
    for _ in tqdm(
        range(n_samples), ascii=True, ncols=100, desc="Sampling", disable=disable_tqdm
    ):
        edges, subtrees, prior_prob = draw_sample_clt(P, False, c=1, coef=10)
        edges_list.append(edges)
        subtrees_list.append(subtrees)
        tree_our_prob_list.append(prior_prob)
    return edges_list, subtrees_list, tree_our_prob_list


def get_samples_info(
    P, my_cell, my_mut, n_samples, subtrees_list=None, disable_tqdm=True
):
    r"""Run some processes on the given samples and returns some raw data.

    If not given first gets the samples.

    If sampling was done internally, it outputs the full sample information.

    If not given_samples:
        O(N * (T_nj * T_obj))
        T_obj = n m^2
        T_nj = n^2 m^2
    Else:
        O(N * n^2)

    :param P: A probability matrix:
            where P_{i,j} is the probability of i\th cell having j\th mutation in
            the latent original matrix.
            In other words, if O is the original matrix and I is the measured
            (observed) matrix then, P_{i,j} = P[O_{i,j}=1|<I_{i,j}, \alpha, \beta>].
    :param my_cell:
    :param my_mut:
    :param n_samples:
    :param subtrees_list: presampled trees. If None new samples will be drawn
    :return:
        if given_samples:   (i.e., subtrees_list is not None)
            pf_cond_list, tree_origin_prob_list
        else:               (i.e., subtrees_list is None)
            pf_cond_list, tree_origin_prob_list, edges_list, subtrees_list,
            tree_our_prob_list
    """
    given_samples = subtrees_list is not None
    if not given_samples:
        edges_list, subtrees_list, tree_our_prob_list = get_samples(P, n_samples)

    pf_cond_list = []
    tree_origin_prob_list = []
    cond_c = np.zeros(P.shape[0], dtype=np.int8)
    cond_c[my_cell] = 1

    for i in tqdm(
        range(n_samples), ascii=True, ncols=100, desc="Sampling", disable=disable_tqdm
    ):
        numerator, denominator = pf_cond_on_one_tree(
            P, subtrees_list[i], cond_c=cond_c, cond_m=my_mut
        )
        pf_cond_list.append(numerator / denominator)

        origin_prob = cell_lineage_tree_prob(P, subtrees_list[i])
        tree_origin_prob_list.append(origin_prob)

    if given_samples:
        return pf_cond_list, tree_origin_prob_list, None, None, None
    else:
        return (
            pf_cond_list,
            tree_origin_prob_list,
            edges_list,
            subtrees_list,
            tree_our_prob_list,
        )


def process_samples(
    pf_cond_list, tree_origin_prob_list, tree_our_prob_list, n_batches=None
):
    """Combine the data corresponding to the sampled trees and output the partition.

    Estimate according to the formula in the paper. One can see this as just
    a simple weighted average.

    Features:
        1. Debiasing is implemented through weights
            (in favour of original probability and against probability of our sampling
        2. Normalization: instead of calculating exact denominator given in the paper.
            We divide the value by summation of the weights.

    :param pf_cond_list:
    :param tree_origin_prob_list:
    :param tree_our_prob_list:
    :param n_batches:
    :return: just one probability through weighted average
    """
    n_samples = len(pf_cond_list)
    estimates = []
    n_batches_internal = n_batches if n_batches is not None else 1
    interval_len = n_samples // n_batches_internal
    for batch_ind in range(n_batches_internal):
        numerator = 0
        denominator = 0
        samples_interval_start = batch_ind * interval_len
        samples_interval_end = min((batch_ind + 1) * interval_len, n_samples)
        assert samples_interval_start < samples_interval_end
        for i in range(samples_interval_start, samples_interval_end):
            numerator += (
                pf_cond_list[i] * tree_origin_prob_list[i] / tree_our_prob_list[i]
            )
            denominator += tree_origin_prob_list[i] / tree_our_prob_list[i]
        estimates.append(numerator / denominator)
    assert len(estimates) >= 1
    if n_batches is None:
        return estimates[0]
    else:
        return estimates



import scphylo as scp


def example(is_expression=False):
    """Return an example for sanity checking and playing with scPhylo.

    is_expression : :obj:`bool`, optional
        Returns the expression dataset instead of the genotype one, by default False

    Returns
    -------
    :class:`anndata.AnnData`
        An object that cells are in `.obs` and mutations are in `.var`.
    """
    if is_expression:
        return scp.io.read(scp.ul.get_file("scphylo.datasets/data/expression.h5ad.gz"))
    else:
        return scp.io.read(scp.ul.get_file("scphylo.datasets/data/genotype.h5ad.gz"))


def test():
    df = scp.io.read(scp.ul.get_file("scphylo.datasets/test/test.tsv"))
    return df


def melanoma20():
    """Mouse Melanoma dataset with 20 sublines.

    This dataset was introduced in :cite:`Wolf_2019` and was used in:

    * :cite:`PhISCS-BnB` Figure 1.

    The size is n_cells × n_muts = 20 × 2367

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.

        `.layers['solution_fig1']` is the solution presented in Figure 1 of PhISCS-BnB
        paper.
    """
    adata = scp.io.read(scp.ul.get_file("scphylo.datasets/real/melanoma20.h5ad"))
    return adata


def colorectal1():
    """Human Colorectal Cancer (Patient 1).

    This dataset was introduced in :cite:`Leung_2017` and was used in:

    * :cite:`B-SCITE` Figure 8a.
    * :cite:`SiFit` Figure 6.
    * :cite:`SPhyR` Table 1.
    * :cite:`SiCloneFit` Figure 3.

    The size is n_cells × n_muts = 178 × 16

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.

    Notes
    -----
    This dataset includes single cells from two sites of the patient body;
    133 single cells from colon as primary tumor site and 45 single cells from liver
    as the tumor metastatic site (178 in total). The number of mutations in this
    dataset is 16. One can remove the cells in this dataset that carry none of these 16
    mutations before feeding it to our network. After removing cells with zero profile,
    the number of cells are 40 and 32 from primary and metastatic sites, respectively
    (72 in total).
    """
    adata = scp.io.read(scp.ul.get_file("scphylo.datasets/real/colorectal1.h5ad"))
    return adata


def colorectal2(readcount=False):
    """Human Colorectal Cancer (Patient 2).

    This dataset was introduced in :cite:`Leung_2017` and was used in:

    * :cite:`PhISCS` Figure 7.
    * :cite:`B-SCITE` Figure 8b.
    * :cite:`SiCloneFit` Figure 4.
    * :cite:`SCARLET` Figure 4.

    The size is n_cells × n_muts = 78 × 25

    Parameters
    ----------
    readcount : :obj:`str`
        Return the readcount information of the data.

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.

            - `.layers['solution_fig7a']` is the solution presented in Figure 7a of
                PhISCS paper.
            - `.layers['solution_fig7b']` is the solution presented in Figure 7b of
                PhISCS paper.
            - `.uns['params_fig7a']` is parameters used as input to get 'solution_7a'.
            - `.uns['params_fig7b']` is parameters used as input to get 'solution_7b'.
            - `.var` includes information of the bulk samples.
    """
    if readcount:
        adata = scp.io.read(
            scp.ul.get_file("scphylo.datasets/real/colorectal2.rc.h5ad")
        )
    else:
        adata = scp.io.read(scp.ul.get_file("scphylo.datasets/real/colorectal2.h5ad"))
    # FIXME: (86 x 25 in B-SCITE) (182 x 36 SiCloneFit)
    # https://github.com/cbg-ethz/infSCITE/blob/master/pat_2.csv
    # https://github.com/hzi-bifo/scelestial-paper-materials-devel/tree/master/testing/
    return adata


def colorectal3():
    """Human Colorectal Cancer.

    This dataset was introduced in :cite:`Wu_2016` and was used in:

    * :cite:`SiFit` Figure 5.

    The size is n_cells × n_muts = 48 × 77

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.
    """
    # adata = scp.io.read(scp.ul.get_file("scphylo.datasets/real/colorectal3.h5ad"))
    # TODO: extract
    return None


def acute_lymphocytic_leukemia1():
    """Human Acute Lymphocytic Leukemia dataset (Patient 1).

    This dataset was introduced in :cite:`Gawad_2014` and was used in:

    * :cite:`B-SCITE` Figure 5.
    * :cite:`infSCITE` Figure S16.

    The size is n_cells × n_muts = 111 × 20

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.
    """
    adata = scp.io.read(
        scp.ul.get_file("scphylo.datasets/real/acute_lymphocytic_leukemia1.h5ad")
    )
    return adata


def acute_lymphocytic_leukemia2():
    """Human Acute Lymphocytic Leukemia dataset (Patient 2).

    This dataset was introduced in :cite:`Gawad_2014` and was used in:

    * :cite:`PhISCS` Figure 9.
    * :cite:`B-SCITE` in Figure 6.
    * :cite:`infSCITE` Figure S17.
    * :cite:`Phyolin` Table 2.

    The size is n_cells × n_muts = 102 × 16

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.

            - `.layers['solution_fig9']` is the solution presented in Figure 9 of PhISCS
                paper.
            - `.uns['params_fig9']` is parameters used as input to get 'solution_fig9'.
            - `.var` includes information of the bulk samples.
    """
    adata = scp.io.read(
        scp.ul.get_file("scphylo.datasets/real/acute_lymphocytic_leukemia2.h5ad")
    )
    # FIXME: 115 x 16 in B-SCITE?
    return adata


def acute_lymphocytic_leukemia3():
    """Human Acute Lymphocytic Leukemia dataset (Patient 3).

    This dataset was introduced in :cite:`Gawad_2014` and was used in:

    * :cite:`infSCITE` Figure S18.
    * :cite:`SCIPhI` Figure 5.
    * :cite:`ScisTree` Figure S3.

    The size is n_cells × n_muts = 150 × 49

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.
    """
    adata = scp.io.read(
        scp.ul.get_file("scphylo.datasets/real/acute_lymphocytic_leukemia3.h5ad")
    )
    # FIXME: 255 x 49 in SCIPhI and ScisTree?
    return adata


def acute_lymphocytic_leukemia4():
    """Human Acute Lymphocytic Leukemia dataset (Patient 4).

    This dataset was introduced in :cite:`Gawad_2014` and was used in:

    * :cite:`infSCITE` Figure S19.
    * :cite:`gpps` Figure 3.
    * :cite:`SASC` Figure 6.

    The size is n_cells × n_muts = 143 × 78

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.
    """
    adata = scp.io.read(
        scp.ul.get_file("scphylo.datasets/real/acute_lymphocytic_leukemia4.h5ad")
    )
    return adata


def acute_lymphocytic_leukemia5():
    """Human Acute Lymphocytic Leukemia dataset (Patient 5).

    This dataset was introduced in :cite:`Gawad_2014` and was used in:

    * :cite:`infSCITE` Figure S20.
    * :cite:`SASC` Figure 7.

    The size is n_cells × n_muts = 96 × 105

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.
    """
    adata = scp.io.read(
        scp.ul.get_file("scphylo.datasets/real/acute_lymphocytic_leukemia5.h5ad")
    )
    return adata


def acute_lymphocytic_leukemia6():
    """Human Acute Lymphocytic Leukemia dataset (Patient 6).

    This dataset was introduced in :cite:`Gawad_2014` and was used in:

    * :cite:`infSCITE` Figure S21.
    * :cite:`Phyolin` Table 2.

    The size is n_cells × n_muts = 146 × 10

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.
    """
    adata = scp.io.read(
        scp.ul.get_file("scphylo.datasets/real/acute_lymphocytic_leukemia6.h5ad")
    )
    return adata


def tnbc():
    """Triple-negative Breast Cancer.

    This dataset was introduced in :cite:`Wang_2014` and was used in:

    * :cite:`SCIPhI` Figure 4.
    * :cite:`B-SCITE` Figure 7.
    * :cite:`TRaIT` Figure 6.

    The size is n_cells × n_muts = 16 × 20

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.

    Examples
    --------
    >>> adata = scp.datasets.tnbc()
    >>> df_in = adata.to_df()
    """
    adata = scp.io.read(scp.ul.get_file("scphylo.datasets/real/tnbc.h5ad"))
    return adata


def erbc():
    """Oestrogen-receptor-positive (ER+) Breast Cancer.

    This dataset was introduced in :cite:`Wang_2014` and was used in:

    * :cite:`SCITE` Figure S8 and S9.
    * :cite:`infSCITE` Figure S15.
    * :cite:`gpps` Figure 1.

    The size is n_cells × n_muts = 47 × 40

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.

            - `.uns['params_scite']` is parameters inferred by SCITE.

    Examples
    --------
    >>> adata = scp.datasets.erbc()
    >>> df_in = adata.to_df()
    """
    adata = scp.io.read(scp.ul.get_file("scphylo.datasets/real/erbc.h5ad"))
    return adata


def muscle_invasive_bladder():
    """Muscle Invasive Bladder Cancer.

    This dataset was introduced in :cite:`Li_2012` and was used in:

    * :cite:`OncoNEM` Figure 6B.

    The size is n_cells × n_muts = 44 × 443

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.

            - `.uns['params_onconem']` is parameters inferred by OncoNEM.

    Examples
    --------
    >>> adata = scp.datasets.muscle_invasive_bladder()
    >>> df_in = adata.to_df()
    """
    adata = scp.io.read(
        scp.ul.get_file("scphylo.datasets/real/muscle_invasive_bladder.h5ad")
    )
    return adata


def renal_cell_carcinoma():
    """Clear-cell Renal-cell Carcinoma.

    This dataset was introduced in :cite:`Xu_2012` and was used in:

    * :cite:`SCITE` Figure S6 and S7.
    * :cite:`infSCITE` Figure S14.

    The size is n_cells × n_muts = 17 × 35

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.

            - `.uns['params_scite']` is parameters inferred by SCITE.

    Examples
    --------
    >>> adata = scp.datasets.renal_cell_carcinoma()
    >>> df_in = adata.to_df()
    """
    adata = scp.io.read(
        scp.ul.get_file("scphylo.datasets/real/renal_cell_carcinoma.h5ad")
    )
    return adata


def myeloproliferative_neoplasms712():
    """JAK2-Negative Myeloproliferative Neoplasm.

    This dataset was introduced in :cite:`Hou_2012` and was used in:

    * :cite:`OncoNEM` Figure 6D.

    The size is n_cells × n_muts = 58 × 712

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.

            - `.uns['params_onconem']` is parameters inferred by OncoNEM.
    """
    adata = scp.io.read(
        scp.ul.get_file("scphylo.datasets/real/myeloproliferative_neoplasms712.h5ad")
    )
    return adata


def myeloproliferative_neoplasms78():
    """JAK2-Negative Myeloproliferative Neoplasm.

    This dataset was introduced in :cite:`Hou_2012` and was used in:

    * :cite:`SCITE` Figure S5.

    The size is n_cells × n_muts = 58 × 78

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.

    Notes
    -----
    The original dataset contains 712 mutations but 78 ones were considered as
    non-synonymous mutations from the full data.
    """
    adata = scp.io.read(
        scp.ul.get_file("scphylo.datasets/real/myeloproliferative_neoplasms78.h5ad")
    )
    return adata


def myeloproliferative_neoplasms18():
    """JAK2-Negative Myeloproliferative Neoplasm.

    This dataset was introduced in :cite:`Hou_2012` and was used in:

    * :cite:`SCITE` Figure S2, S3 and S4.
    * :cite:`Kim_2014` Figure 1.
    * :cite:`infSCITE` Figure S13.
    * :cite:`gpps` Figure 2.

    The size is n_cells × n_muts = 58 × 18

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.

            - `.uns['params_scite']` is parameters inferred by SCITE.

    Notes
    -----
    The original dataset contains 712 mutations but 18 ones were considered as
    cancer related mutations from the full data.
    """
    adata = scp.io.read(
        scp.ul.get_file("scphylo.datasets/real/myeloproliferative_neoplasms18.h5ad")
    )
    return adata


def high_grade_serous_ovarian_cancer1():
    """High Grade Serous Ovarian Cancer (Patient 2).

    This dataset was introduced in :cite:`McPherson_2016` and was used in:

    * :cite:`infSCITE` Figure S22.

    The size is n_cells × n_muts = 588 × 37

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.
    """
    # adata = scp.io.read(
    #     scp.ul.get_file(
    #         "scphylo.datasets/real/high_grade_serous_ovarian_cancer1.h5ad"
    #     )
    # )
    # TODO: extract
    return None


def high_grade_serous_ovarian_cancer2():
    """High Grade Serous Ovarian Cancer (Patient 3).

    This dataset was introduced in :cite:`McPherson_2016` and was used in:

    * :cite:`infSCITE` Figure S23.

    The size is n_cells × n_muts = 672 × 60

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.
    """
    # adata = scp.io.read(
    #     scp.ul.get_file(
    #         "scphylo.datasets/real/high_grade_serous_ovarian_cancer2.h5ad"
    #     )
    # )
    # TODO: extract
    return None


def high_grade_serous_ovarian_cancer3():
    """High Grade Serous Ovarian Cancer (Patient 9).

    This dataset was introduced in :cite:`McPherson_2016` and was used in:

    * :cite:`infSCITE` Figure S24.
    * :cite:`scVILP` Figure 5.
    * :cite:`SCIPhI` Figure S10.

    The size is n_cells × n_muts = 420 × 37

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.
    """
    # adata = scp.io.read(
    #     scp.ul.get_file(
    #         "scphylo.datasets/real/high_grade_serous_ovarian_cancer3.h5ad"
    #     )
    # )
    # TODO: extract
    # FIXME: 370 x 43 in scVILP and SCIPhI?
    return None


def high_grade_serous_ovarian_cancer_3celllines():
    """High Grade Serous Ovarian Cancer (3 cell lines).

    This dataset was introduced in :cite:`Laks_2019`
    and preprocessed in :cite:`McPherson_2019`.
    The phylogeny is presented in Figure 3H.

    The size is n_cells × n_muts = 891 × 13666

    Note that 402 mutations were deleted during evolution
    in the inferred tree by original study So, here they were filtered out
    (meaning they are 14068 mutations in total in the original study).

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy
        (it is obtained based on the number of mutant/total reads).

            - `.obs['clone_id']` is the clone id to which the cell is assigned in
                Figure 3H.
            - `.obs['group_color']` is unique colors for each 'clone_id'.
            - `.obs['cell_name']` is a new name for each cell based on the
                'group_color'.
            - `.layers['mutant']` is the number of mutant reads at each locus in each
                cell.
            - `.layers['total']` is the total number of reads at each locus in each
                cell.
            - `.layers['ground']` is the solution inferred in Figure 3H of the original
                paper.
            - `.uns['params_ground']` is parameters inferred by comparing ground and
                noisy matrices.
            - `.var` includes information of the bulk samples.

    Examples
    --------
    >>> sc = scp.datasets.high_grade_serous_ovarian_cancer_3celllines()
    >>> print(sc)
    AnnData object with n_obs × n_vars = 891 × 13666
        obs: 'clone_id', 'group_color', 'cell_name'
        uns: 'params_ground'
        layers: 'ground', 'mutant', 'total'
    """
    adata = scp.io.read(scp.ul.get_file("scphylo.datasets/real/ovarian.h5ad.gz"))
    adata.var_names = adata.var_names.str.replace(":", "_")
    return adata


def oligodendroglioma_idh_mutated_tumor():
    """Oligodendroglioma IDH-mutated tumor.

    This dataset was introduced in :cite:`Tirosh_2016` and was used in:

    * :cite:`SASC` Figure 5.

    The size is n_cells × n_muts = 579 × 77

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.
    """
    adata = scp.io.read(
        scp.ul.get_file(
            "scphylo.datasets/real/oligodendroglioma_idh_mutated_tumor.h5ad"
        )
    )
    return adata


def acute_lymphocytic_leukemia_many():
    """Human Acute Lymphocytic Leukemia datasets.

    This dataset was introduced in :cite:`Morita_2020` and was used in:

    * :cite:`Phyolin` Figure 3 and Table 1.

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.
    """
    # TODO: extract all
    return None


def isogenic_fibroblast_cell_line():
    """Isogenic Fibroblast cell line dataset.

    This dataset was introduced in :cite:`Leung_2015` and was used in:

    * :cite:`SCIPhI` Figure S9.

    The size is n_cells × n_muts = 19 × ?

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.
    """
    # TODO: extract
    return None


def acute_myeloid_leukemia1():
    """Human Acute Myeloid Leukemia dataset (Patient 1).

    This dataset was introduced in :cite:`Velten_2021` and was used in:

    The size is n_cells × n_muts = 1430 × 15

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.
    """
    adata = scp.io.read(
        scp.ul.get_file("scphylo.datasets/real/acute_myeloid_leukemia1.h5ad")
    )
    return adata


def acute_myeloid_leukemia2():
    """Human Acute Myeloid Leukemia dataset (Patient 2).

    This dataset was introduced in :cite:`Velten_2021` and was used in:

    The size is n_cells × n_muts = 1066 × 21

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.
    """
    adata = scp.io.read(
        scp.ul.get_file("scphylo.datasets/real/acute_myeloid_leukemia2.h5ad")
    )
    return adata



"""Command Line Interface."""

import click

import scphylo as scp


@click.group()
@click.version_option()
def cli():
    """scphylo-tools command line interface."""
    pass


@cli.group()
def solver():
    """Solver commands."""
    pass


@cli.group()
def utils():
    """Run utility commands."""
    pass


@solver.command()
@click.argument("data")
@click.argument("alpha", type=float)
@click.argument("beta", type=float)
def scistree(data, alpha, beta):
    """Run ScisTree solver."""
    df = scp.io.read(data)
    scp.tl.scistree(df, alpha=alpha, beta=beta)


@solver.command()
@click.argument("data")
@click.argument("alpha", type=float)
@click.argument("beta", type=float)
def huntress(data, alpha, beta):
    """Run HUNTRESS solver."""
    df = scp.io.read(data)
    scp.tl.huntress(df, alpha=alpha, beta=beta)


@solver.command()
@click.argument("data")
@click.argument("alpha", type=float)
@click.argument("beta", type=float)
@click.option("-r", "--n_restarts", default=3, type=int)
@click.option("-l", "--n_iters", default=1000, type=int)
@click.option("-e", "--experiment", is_flag=True)
@click.option("-t", "--n_threads", default=1, type=int)
@click.option("-s", "--seed", default=1, type=int)
def scite(data, alpha, beta, n_restarts, n_iters, experiment, n_threads, seed):
    """Run SCITE solver."""
    df = scp.io.read(data)
    scp.tl.scite(df, alpha=alpha, beta=beta, n_restarts=n_restarts, n_iters=n_iters)


@solver.command()
@click.argument("data")
@click.argument("alpha", type=float)
@click.argument("beta", type=float)
def phiscsb(data, alpha, beta):
    """Run PhISCS-B solver."""
    df = scp.io.read(data)
    scp.tl.phiscsb(df, alpha=alpha, beta=beta)


@solver.command()
@click.argument("data")
@click.argument("alpha", type=float)
@click.argument("beta", type=float)
@click.option("--solver", default="scite")
@click.option("--n_samples", default=100, type=int)
@click.option("--sample_size", default=15, type=int)
@click.option("--n_jobs", default=1, type=int)
@click.option("--n_iterations", default=10000, type=int)
def booster(data, alpha, beta, solver, n_samples, sample_size, n_jobs, n_iterations):
    """Run Booster solver."""
    df = scp.io.read(data)
    scp.tl.booster(
        df,
        alpha=alpha,
        beta=beta,
        solver=solver,
        n_samples=n_samples,
        sample_size=sample_size,
        n_jobs=n_jobs,
        n_iterations=n_iterations,
    )


@solver.command()
@click.argument("data")
@click.option("-b", "--bounding", default="simulated")
def bnb(data, bounding):
    """Run BnB solver."""
    df = scp.io.read(data)
    scp.tl.bnb(df, bounding=bounding)


@utils.command()
@click.argument("file1")
@click.argument("file2")
@click.argument("output")
def consensus(file1, file2, output):
    """Consensus tree."""
    sc1 = scp.io.read(file1)
    sc2 = scp.io.read(file2)
    tree = scp.tl.consensus(sc1, sc2)
    df = scp.ul.to_cfmatrix(tree)
    scp.io.write(df, output)


@utils.command()
@click.argument("file")
def cf2newick(file):
    """Convert CFMatrix to Newick."""
    pass


@utils.command()
@click.argument("file")
@click.option("-p", "--processes", default=1, type=int)
def search(file, processes):
    """Search."""
    pass


@utils.command()
@click.argument("file1")
@click.argument("file2")
def score(file1, file2):
    """Score."""
    pass


@utils.command()
@click.argument("file")
def cf2tree(file):
    """Convert CFMatrix to Tree."""
    pass


@utils.command()
@click.argument("data")
@click.argument("alpha", type=float)
@click.argument("beta", type=float)
@click.option("--n_threads", default=1, type=int)
@click.option("--n_samples", default=100, type=int)
def partf(data, alpha, beta, n_threads, n_samples):
    """Partition function."""
    df = scp.io.read(data)
    scp.tl.partition_function(df, alpha=alpha, beta=beta, n_samples=n_samples)



import os
import time

import networkx as nx

import scphylo as scp


def grmt(df_input, alpha, beta, n_iters=30, n_threads=1):
    """Solving using GRMT.

    Generative Reconstruction of Mutation Tree From Scratch Using Single-Cell
    Sequencing Data :cite:`GRMT`.

    Parameters
    ----------
    df_input : :class:`pandas.DataFrame`
        Input genotype matrix in which rows are cells and columns are mutations.
        Values inside this matrix show the presence (1), absence (0) and missing
        entires (3).
    alpha : :obj:`float`
        False positive error rate.
    beta : :obj:`float`
        False negative error rate.
    n_iters : :obj:`int`, optional
        Number of iterations, by default 30
    n_threads : :obj:`int`, optional
        Number of threads, by default 1

    Returns
    -------
    :class:`pandas.DataFrame`
        A conflict-free matrix in which rows are cells and columns are mutations.
        Values inside this matrix show the presence (1) and absence (0).
    """
    executable = scp.ul.executable("grmt", "GRMT")

    scp.logg.info(
        f"running GRMT with alpha={alpha}, beta={beta}, n_threads={n_threads}"
    )

    tmpdir = scp.ul.tmpdirsys(suffix=".sphyr")

    df_input.to_csv(f"{tmpdir.name}/grmt.input", sep="\t", header=None, index=None)
    with open(f"{tmpdir.name}/grmt.cellnames", "w") as fout:
        fout.write("\n".join(df_input.index) + "\n")
    with open(f"{tmpdir.name}/grmt.mutnames", "w") as fout:
        fout.write("\n".join(df_input.columns) + "\n")

    cmd = (
        f"{executable} "
        f"--input {tmpdir.name}/grmt.input "
        f"--alpha {alpha} "
        f"--beta {beta} "
        f"--clabel {tmpdir.name}/grmt.cellnames "
        f"--mlabel {tmpdir.name}/grmt.mutnames "
        "--lambda 0.7 "
        "--kappa 1 "
        "--n_init 100 "
        f"--n_iter {n_iters} "
        "--maxl 0 "
        f"--threads {n_threads} "
        f"--output {tmpdir.name}/grmt "
        f"&> {tmpdir.name}/grmt.log"
    )

    s_time = time.time()
    os.system(cmd)
    e_time = time.time()
    running_time = e_time - s_time

    df_output = df_input.copy()
    df_output[:] = 0
    G = nx.drawing.nx_pydot.read_dot(f"{tmpdir.name}/grmt.dot")
    relation = {
        v.replace('"', "").replace("\\n", ""): k
        for k, v in nx.get_node_attributes(G, "label").items()
    }
    data = {}
    for k, v in relation.items():
        if " " in k:
            for t in k.split(" "):
                data[t] = v
        else:
            data[k] = v
    for cell in df_output.index:
        muts = nx.shortest_path(G, source="0", target=data[cell])
        muts.remove("0")
        muts.remove(data[cell])
        muts = [G.nodes[m]["label"].replace('"', "") for m in muts]
        if len(muts) > 0:
            df_output.loc[cell, muts] = 1

    tmpdir.cleanup()

    scp.ul.stat(df_input, df_output, alpha, beta, running_time)

    return df_output



import pandas as pd
from IPython.display import Image, display

import scphylo as scp


def infercna(expr, ref_cells, genome="hg19", show_heatmap=False):
    """Call copy number profiles from scRNAseq.

    This function is a wrapper to run
    `InferCNA <https://github.com/jlaffy/infercna>`_

    Parameters
    ----------
    expr : :class:`anndata.AnnData`
        The input expression matrix where

            - raw count are in `.X`
            - fpkm is in `.layer['fpkm']`
            - tpm is in `.layer['tpm']`
    ref_cells : :obj:`dict`
        A dictionary of list of normal cells.
    genome : :obj:`str`, optional
        {'hg38', 'hg19', 'mm10'}, by default "hg19"
    show_heatmap : :obj:`bool`, optional
        Show the copy number heatmap, by default False

    Returns
    -------
    :class:`pandas.DataFrame`
        A datafrme of log2 copy number ratios where normal cells
        are filtered out. Rows are cells and columns are list of
        variable genes.

    Examples
    --------
    >>> cna_log = scp.tl.infercna(crc1, {'normal': ['NC_502']})
    """
    infercna, infercna_is_not_imported = scp.ul.import_rpy2(
        "infercna",
        "devtools::install_github(c('jlaffy/scalop','jlaffy/infercna'))\n",
    )
    if infercna_is_not_imported:
        scp.logg.error("Unable to import a package!")

    import rpy2.robjects as ro
    from rpy2.robjects import pandas2ri
    from rpy2.robjects.lib import grdevices
    from rpy2.robjects.packages import importr

    base = importr("base")

    scp.logg.info("running inferCNA")

    rc = {key: ro.StrVector(value) for key, value in ref_cells.items()}
    rc = ro.ListVector(rc)

    tpm = expr.to_df(layer="tpm").T
    # tpm.index = tpm.index.str.split("_").str[1]

    with ro.conversion.localconverter(ro.default_converter + pandas2ri.converter):
        data = ro.conversion.py2rpy(tpm)
    data = base.as_matrix(data)
    data.rownames = ro.StrVector(tpm.index.tolist())

    infercna.useGenome(genome)
    # ref = infercna.retrieveGenome()

    cna_mal = infercna.infercna(
        data,
        refCells=rc,
        window=100,
        n=5000,
        range=ro.FloatVector([-3, 3]),
        noise=0,
        center_method="median",
        isLog=False,
        verbose=False,
    )
    ro.globalenv["cna_mal"] = cna_mal
    ro.globalenv["rc"] = rc
    cna_mal = ro.r("cna_mal[, !colnames(cna_mal) %in% unlist(rc)]")

    cna_plot = infercna.cnaPlot(
        cna_mal,
        limits=ro.FloatVector([-0.5, 0.5]),
        cols=infercna.heatCols,
        ratio=0.5,
        x_name="Chromosome",
        y_name="Cell",
        legend_title="Inferred CNA\n[log2 ratio]",
        x_hide=ro.StrVector(["13", "18", "21", "Y"]),
        order_cells=False,
        subset_genes=ro.NULL,
        euclid_dist=False,
        angle=ro.NULL,
        x_angle=ro.NULL,
        y_angle=0,
        axis_rel=1,
        base_size=12,
        axis_title_size=12,
        axis_text_size=11,
        base_col="#073642",
        title=ro.NULL,
        subtitle=ro.NULL,
        caption=ro.NULL,
        text_size=12,
        y_hide=ro.NULL,
        tile_size=0.1,
        tile_col=ro.NULL,
        legend_position="right",
        legend_height=2,
        legend_width=0.6,
        legend_rel=0.9,
        legend_colour="black",
        legend_breaks=ro.NULL,
        legend_labels=ro.NULL,
        legend_justification="top",
        legend_title_position="bottom",
        legend_title_angle=ro.NULL,
        legend_title_rel=0.9,
    )

    # clones = infercna.findClones(
    #     cna_mal,
    #     prob=0.95,
    #     coverage=0.8,
    #     mode_size=10,
    #     clone_size=3,
    #     by="arm",
    #     bySampling=False,
    #     nsamp=2000,
    #     force_tries=False,
    #     verbose=False,
    # )
    # clones = {key: clones.rx2(key) for key in clones.names}

    # cor = infercna.refCorrect(cna=cna_mal, noise=ro.NULL, isLog=True)
    # print(cor)

    with ro.conversion.localconverter(ro.default_converter + pandas2ri.converter):
        cna_log = ro.conversion.rpy2py(cna_plot.rx2("data"))
    cna_log = pd.pivot(cna_log, index="Cell", columns="Gene", values="CNA")

    if show_heatmap:
        with ro.lib.grdevices.render_to_bytesio(
            grdevices.png, width=1024, height=896, res=150
        ) as image:
            ro.r.show(cna_plot.rx2("p"))
        display(Image(data=image.getvalue(), embed=True, retina=True))

    return cna_log



import datetime
import time

import numpy as np
import pandas as pd
from joblib import Parallel, delayed

import scphylo as scp
from scphylo.tl.partition_function._pf import (
    get_samples,
    get_samples_info,
    process_samples,
)


def partition_function(df_input, alpha, beta, n_samples, n_batches, muts, cells):
    """Calculate the probability of a mutation seeding particular cells.

    Parameters
    ----------
    df_input : :class:`pandas.DataFrame`
        Input genotype matrix.
    alpha : float
        False positive error rate.
    beta : float
        False negative error rate.
    n_samples : int
        Number of samples to get from the distribution (suggest: 1000)
    n_batches : int
        Number of batches to repeat the experiment (suggest: 100)
    muts : list
        The list of mutations
    cells : list
        The list of cells

    Returns
    -------
    :class:`pandas.DataFrame`
        A table of probabilities for every mutation and every batch.
    """
    df_output = pd.DataFrame(None, index=muts, columns=range(n_batches))
    s_time = time.time()
    I_mtr = df_input.values
    t1 = I_mtr * (1 - beta) / (alpha + 1 - beta)
    t2 = (1 - I_mtr) * beta / (beta + 1 - alpha)
    P = t1 + t2
    P[I_mtr == 3] = 0.5

    my_muts = np.where(df_input.columns.isin(muts))[0]
    my_cells = np.where(df_input.index.isin(cells))[0]
    if len(my_muts) != len(muts):
        scp.logg.error("bad muts choise!")
    if len(my_cells) != len(cells):
        scp.logg.error("bad cells choise!")

    _, subtrees_list, tree_our_prob_list = get_samples(P, n_samples)

    def run(mut):
        my_mut = np.where(df_input.columns == mut)[0][0]
        pf_cond_list, tree_origin_prob_list, _, _, _ = get_samples_info(
            P, my_cells, my_mut, n_samples, subtrees_list
        )
        estimates = process_samples(
            pf_cond_list, tree_origin_prob_list, tree_our_prob_list, n_batches
        )
        return mut, estimates

    output = Parallel(n_jobs=len(muts))(delayed(run)(mut) for mut in muts)
    for mut, estimates in output:
        df_output.loc[mut] = estimates

    e_time = time.time()
    running_time = e_time - s_time
    scp.logg.info(f"elapsed time: {datetime.timedelta(seconds=running_time)}")

    return df_output.astype(float)



import os

import click

import scphylo as scp
from scphylo.pl._trees import _newick_info2_mutation_list


@click.command(short_help="Convert conflict-free to newick file.")
@click.argument(
    "cf_file",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
def cf2newick(cf_file):
    """Convert conflict-free matrix to newick."""
    outfile = os.path.splitext(cf_file)[0]
    cf = scp.io.read(cf_file)
    tree = scp.ul.to_tree(cf)
    newick, info2, mutations = _newick_info2_mutation_list(tree)
    with open(f"{outfile}.newick", "w") as fout:
        fout.write(newick + "\n")
    info2.to_csv(f"{outfile}.info2", index=None)

    return None


@click.command(short_help="Convert conflict-free to clonal tree.")
@click.argument(
    "cf_file",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
def cf2tree(cf_file):
    """Draw conflict-free matrix to clonal tree."""
    outfile = os.path.splitext(cf_file)[0]
    cf = scp.io.read(cf_file)
    tree = scp.ul.to_tree(cf)
    scp.pl.clonal_tree(tree, output_file=f"{outfile}.png")

    return None



from joblib import Parallel, delayed
from tqdm import tqdm

import scphylo as scp


def subsampling(
    df_input,
    alpha,
    beta,
    solver,
    sample_on,
    sample_size,
    n_samples,
    begin_sample,
    n_jobs,
    time_limit,
    n_iterations,
    tmpdir,
    disable_tqdm,
):
    if solver.lower() == "scite":
        time_limit = n_iterations

    @scp.ul.with_timeout(time_limit)
    def run(i):
        if sample_on == "muts":
            dfn = df_input.sample(n=sample_size, replace=False, axis=1)
        elif sample_on == "cells":
            dfn = df_input.sample(n=sample_size, replace=False, axis=0)
        else:
            scp.logg.error("`sample_on` must be either 'muts' or 'cells'!")

        if solver.lower() == "phiscs":
            dfo = scp.tl.phiscsb(dfn, alpha, beta, experiment=True)
            dfo.to_csv(f"{tmpdir}/{i}.CFMatrix", sep="\t")
        elif solver.lower() == "scite":
            dfo, _, _, _ = scp.tl.scite(
                dfn, alpha, beta, n_iters=n_iterations, n_restarts=1, experiment=True
            )
            dfo.to_csv(f"{tmpdir}/{i}.CFMatrix", sep="\t")
        elif solver.lower() == "scistree":
            x = ((dfn == 1).sum() >= 2) & ((dfn != 3).sum() >= 0.2 * dfn.shape[0])
            dfn = dfn[dfn.columns[x]]
            if dfn.shape[1] < 2:
                return None
            dfo, _ = scp.tl.scistree(dfn, alpha, beta, experiment=True)
            dfo.to_csv(f"{tmpdir}/{i}.CFMatrix", sep="\t")
        else:
            scp.logg.error("`solver` must be either 'phiscs' or 'scite'!")

    with scp.ul.tqdm_joblib(
        tqdm(
            ascii=True,
            ncols=100,
            desc="SUBSAMPLING    (1/3)",
            total=n_samples,
            position=0,
            disable=disable_tqdm,
        )
    ):
        Parallel(n_jobs=n_jobs)(
            delayed(run)(i) for i in range(begin_sample, begin_sample + n_samples)
        )
